#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# State
STATE_NEW = "NEW"
STATE_RUNNING = "RUNNING"
STATE_TIMEOUT = "TIMEOUT"
STATE_STOPPED = "STOPPED"
STATE_SHUTDOWN = "SHUTDOWN"

# IOC Scopes
PROTOTYPE = "prototype"
SINGLETON = "singleton"
